#region License Information
/* HeuristicLab
 * Copyright (C) Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Parameters;
using HEAL.Attic;
using System;
using HeuristicLab.Data;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Regression.Extensions {
  [Item("TrainingBestSolutionAnalyzer (with constraints)", "An operator that analyzes the training best symbolic regression solution for single objective symbolic regression problems.")]
  [StorableType("93A9331C-9E50-45DE-804B-21785A07EFB4")]
  public sealed class TrainingBestSolutionAnalyzer : SymbolicDataAnalysisSingleObjectiveTrainingBestSolutionAnalyzer<ISymbolicRegressionSolution>,
  ISymbolicDataAnalysisInterpreterOperator, ISymbolicDataAnalysisBoundedOperator {
    private const string ProblemDataParameterName = "ProblemData";
    private const string SymbolicDataAnalysisTreeInterpreterParameterName = "SymbolicDataAnalysisTreeInterpreter";
    private const string EstimationLimitsParameterName = "EstimationLimits";
    #region parameter properties
    public ILookupParameter<IRegressionProblemData> ProblemDataParameter {
      get { return (ILookupParameter<IRegressionProblemData>)Parameters[ProblemDataParameterName]; }
    }
    public ILookupParameter<ISymbolicDataAnalysisExpressionTreeInterpreter> SymbolicDataAnalysisTreeInterpreterParameter {
      get { return (ILookupParameter<ISymbolicDataAnalysisExpressionTreeInterpreter>)Parameters[SymbolicDataAnalysisTreeInterpreterParameterName]; }
    }
    public IValueLookupParameter<DoubleLimit> EstimationLimitsParameter {
      get { return (IValueLookupParameter<DoubleLimit>)Parameters[EstimationLimitsParameterName]; }
    }
    #endregion

    [StorableConstructor]
    private TrainingBestSolutionAnalyzer(StorableConstructorFlag _) : base(_) { }
    private TrainingBestSolutionAnalyzer(TrainingBestSolutionAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public TrainingBestSolutionAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<IRegressionProblemData>(ProblemDataParameterName, "The problem data for the symbolic regression solution."));
      Parameters.Add(new LookupParameter<ISymbolicDataAnalysisExpressionTreeInterpreter>(SymbolicDataAnalysisTreeInterpreterParameterName, "The symbolic data analysis tree interpreter for the symbolic expression tree."));
      Parameters.Add(new ValueLookupParameter<DoubleLimit>(EstimationLimitsParameterName, "The lower and upper limit for the estimated values produced by the symbolic regression model."));
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new TrainingBestSolutionAnalyzer(this, cloner);
    }

    protected override ISymbolicRegressionSolution CreateSolution(ISymbolicExpressionTree bestTree, double bestQuality) {
      if (!ApplyLinearScalingParameter.ActualValue.Value) throw new NotSupportedException("This analyzer only works if linear scaling of models is activated.");

      var problemData = ProblemDataParameter.ActualValue;
      var solTree = (ISymbolicExpressionTree)bestTree.Clone();
      using (var nls = new ConstrainedNLSInternal("MMA", solTree, 100, ProblemDataParameter.ActualValue)) {
        var originalConstraintValues = (double[])nls.BestConstraintValues.Clone(); // for debugging
        nls.Optimize(ConstrainedNLSInternal.OptimizationMode.UpdateParametersAndKeepLinearScaling);

        var model = new SymbolicRegressionModel(problemData.TargetVariable, solTree,
          SymbolicDataAnalysisTreeInterpreterParameter.ActualValue, EstimationLimitsParameter.ActualValue.Lower, EstimationLimitsParameter.ActualValue.Upper);
        var sol = new SymbolicRegressionSolution(model, (IRegressionProblemData)problemData.Clone());
        // debugging
        sol.AddOrUpdateResult("Constraint values (after optimization)", new DoubleArray(nls.BestConstraintValues));
        sol.AddOrUpdateResult("Constraint values (before optimization)", new DoubleArray(originalConstraintValues));
        sol.AddOrUpdateResult("Quality before optimization in analyzer", new DoubleValue(bestQuality));
        sol.AddOrUpdateResult("Quality after optimization in analyzer", new DoubleValue(nls.BestError));
        sol.AddOrUpdateResult("NLOpt result", new StringValue(nls.OptResult.ToString()));
        return sol;
      }
    }
  }
}
